import numpy as np

from .definitions import APPROACH


def folded_log_formulation(duration, param, case, param_mean=None, duration_mean=None):
    """

    Args:
        duration:
        param:
        case:
        param_mean:
        duration_mean:

    Returns:

    """
    if param_mean and duration_mean:
        mean_ln_duration = np.log(duration_mean)
        mean_ln_param = np.log(param_mean)
    else:
        param_mean = param.mean()
        mean_ln_param = np.log(param).mean()
        mean_ln_duration = np.log(duration).mean()

    divisor = ((np.log(duration) - mean_ln_duration) ** 2).sum()

    if case == APPROACH.LOG2:
        # for the twofold formulation
        b = ((np.log(param) - mean_ln_param) * (np.log(duration) - mean_ln_duration)).sum() / divisor
        a = mean_ln_param - b * mean_ln_duration

    elif case == APPROACH.LOG1:
        # for the onefold formulation
        b = ((param - param_mean) * (np.log(duration) - mean_ln_duration)).sum() / divisor
        a = param_mean - b * mean_ln_duration

    else:
        raise NotImplementedError

    return float(a), float(b)


def hyperbolic_formulation(duration, param, a_start=20.0, b_start=15.0, param_mean=None, duration_mean=None):
    """

    Args:
        duration:
        param:
        a_start:
        b_start:
        param_mean:
        duration_mean:

    Returns:

    """

    # ------------------------------------------------------------------------------------------------------------------
    def get_param(dur, par, a_, b_):
        i = -a_ / (dur + b_)
        if param_mean:
            i_mean = - param_mean / duration_mean
            param_mean_ = param_mean
        else:
            i_mean = i.mean()
            param_mean_ = par.mean()

        b_ = ((par - param_mean_) * (i - i_mean)).sum() / ((i - i_mean) ** 2).sum()
        a_ = param_mean_ - b_ * i_mean
        return a_, b_

    # ------------------------------------------------------------------------------------------------------------------
    iteration_steps = 0

    a = a_start
    b = b_start

    conditions = True
    while conditions:
        conditions = False

        a_s = a
        b_s = b
        a, b = get_param(duration, param, a, b)
        conditions = (abs(a - a_s) > 0.005) or (abs(b - b_s) > 0.005) or conditions
        a = (a + a_s) / 2
        b = (b + b_s) / 2

        iteration_steps += 1
    return float(a), float(b)



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import warnings
from math import floor
from os import path, mkdir
from webbrowser import open as show_file
from scipy.optimize import newton

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import pandas as pd
from tqdm import tqdm

from .arg_parser import heavy_rain_parser
from .idf_backend import IdfParameters
from .little_helpers import minutes_readable, height2rate, delta2min, rate2height, frame_looper, event_caption
from .definitions import *
from .in_out import import_series
from .sww_utils import (remove_timezone, guess_freq, rain_events, agg_events, event_duration, resample_rain_series,
                        rain_bar_plot, IdfError, )
from .plot_helpers import idf_bar_axes


########################################################################################################################
class IntensityDurationFrequencyAnalyse:
    """
    heavy rain as a function of the duration and the return period acc. to DWA-A 531 (2012)

    This program reads the measurement data of the rainfall
    and calculates the distribution of the rainfall as a function of the return period and the duration
    
    for duration steps up to 12 hours (and more) and return period in a range of '0.5a <= T_n <= 100a'
    """

    def __init__(self, series_kind=SERIES.PARTIAL, worksheet=METHOD.KOSTRA, extended_durations=False):
        """
        heavy rain as a function of the duration and the return period acc. to DWA-A 531 (2012)

        This program reads the measurement data of the rainfall
        and calculates the distribution of the rainfall as a function of the return period and the duration

        for duration steps up to 12 hours (and more) and return period in a range of '0.5a <= T_n <= 100a'

        Args:
            series_kind (str): ['partial', 'annual']
            worksheet (str): ['DWA-A_531', 'ATV-A_121', 'DWA-A_531_advektiv']
            extended_durations (bool): add [720, 1080, 1440, 2880, 4320, 5760, 7200, 8640] minutes to the calculation
        """
        self._series = None  # type: pd.Series # rain time-series
        self._freq = None  # frequency of the rain series

        #  how to calculate the idf curves
        # IdfParameters.__init__(self, series_kind=series_kind, worksheet=worksheet,
        #                                 extended_durations=extended_durations)
        self._parameter = IdfParameters(series_kind=series_kind, worksheet=worksheet,
                                        extended_durations=extended_durations)
        # self._parameter = None  # type: IdfParameters #  how to calculate the idf curves

        self._return_periods_frame = None  # type: pd.DataFrame # with return periods of all given durations
        self._rain_events = None
        self._rainfall_sum_frame = None  # type: pd.DataFrame # with rain sums of all given durations

    # __________________________________________________________________________________________________________________
    @property
    def series(self):
        if self._series is None:
            raise IdfError('No Series defined for IDF-Analysis!')
        return self._series

    @series.setter
    def series(self, series):
        self._series = series

    def set_series(self, series):
        """
        set the series for the analysis

        Args:
            series (pandas.Series): precipitation time-series
        """
        if not isinstance(series, pd.Series):
            raise IdfError('The series has to be a pandas Series.')

        if not isinstance(series.index, pd.DatetimeIndex):
            raise IdfError('The series has to have a DatetimeIndex.')

        if series.index.tz is not None:
            series = remove_timezone(series)

        self._freq = guess_freq(series.index)
        freq_minutes = delta2min(self._freq)
        self._parameter.filter_durations(freq_minutes)
        self.series = series.replace(0, np.NaN).dropna()
        self._return_periods_frame = None
        self._rain_events = None
        self._rainfall_sum_frame = None

    # __________________________________________________________________________________________________________________
    @property
    def duration_steps(self):
        """
        get duration steps (in minutes) for the parameter calculation and basic evaluations
        Returns:
            list | numpy.ndarray: duration steps in minutes
        """
        # if not self.parameters:
        #     raise IdfError('No Series defined for IDF-Analysis!')
        return self.parameters.durations

    @duration_steps.setter
    def duration_steps(self, durations):
        """
        set duration steps (in minutes) for the parameter calculation and basic evaluations
        Args:
            durations (list | numpy.ndarray): duration steps in minutes
        """
        if not isinstance(durations, (list, np.ndarray)):
            raise IdfError('Duration steps have to be {} got "{}"'.format((list, np.ndarray), type(durations)))
        self.parameters.durations = durations

    # __________________________________________________________________________________________________________________
    @property
    def parameters(self):
        """
        get the calculation parameters

        calculation method depending on the used worksheet and on the duration
        also the parameters for each method

        to save some time and save the parameters with
        :func:`IntensityDurationFrequencyAnalyse.write_parameters`
        and read them later with :func:`IntensityDurationFrequencyAnalyse.read_parameters`

        Returns:
            IdfParameters: calculation parameters
        """
        if not self._parameter:
            self._parameter.calc_from_series(self.series)
        return self._parameter

    def write_parameters(self, filename):
        """
        save parameters as yaml-file to save computation time.

        Args:
            filename (str): filename for the parameters yaml-file
        """
        self.parameters.to_yaml(filename)

    def read_parameters(self, filename):
        """
        read parameters from a .yaml-file to save computation time.
        extract interim results from parameters

        Args:
            filename (str): filename of the parameters yaml-file
        """
        self._parameter = IdfParameters.from_yaml(filename)

    def auto_save_parameters(self, filename):
        """auto-save the parameters as a yaml-file to save computation time."""
        if path.isfile(filename):
            self.read_parameters(filename)
        else:
            from os import makedirs
            parent = path.dirname(filename)
            if not path.isdir(parent):
                makedirs(parent)
            self.write_parameters(filename)

    # __________________________________________________________________________________________________________________
    def depth_of_rainfall(self, duration, return_period):
        """
        calculate the height of the rainfall h in L/m² = mm

        Args:
            duration (int | float | list | numpy.ndarray | pandas.Series): duration: in minutes
            return_period (float): in years

        Returns:
            int | float | list | numpy.ndarray | pandas.Series: height of the rainfall h in L/m² = mm
        """
        if self.parameters.series_kind == SERIES.ANNUAL:
            if return_period < 5:
                print('WARNING: Using an annual series and a return period < 5 a will result in faulty values!')

            if return_period <= 10:
                return_period = np.exp(1.0 / return_period) / (np.exp(1.0 / return_period) - 1.0)

            log_tn = -np.log(np.log(return_period / (return_period - 1.0)))

        else:
            log_tn = np.log(return_period)

        u, w = self.parameters.get_u_w(duration)
        return u + w * log_tn

    # __________________________________________________________________________________________________________________
    def rain_flow_rate(self, duration, return_period):
        """
        convert the height of rainfall to the specific rain flow rate in [l/(s*ha)]
        if 2 array-like parameters are give, a element-wise calculation will be made.
        So the length of the array must be the same.

        Args:
            duration (int | float | list | numpy.ndarray | pandas.Series): in minutes
            return_period (float): in years

        Returns:
                int | float | list | numpy.ndarray | pandas.Series: specific rain flow rate in [l/(s*ha)]
        """
        height_of_rainfall = self.depth_of_rainfall(duration=duration, return_period=return_period)
        return height2rate(height_of_rainfall=height_of_rainfall, duration=duration)

    # __________________________________________________________________________________________________________________
    def r_720_1(self):
        """
        rain flow rate in [l/(s*ha)] for a duration of 12h and a return period of 1 year

        Returns:
            float: rain flow rate in [l/(s*ha)]
        """
        return self.rain_flow_rate(duration=720, return_period=1)

    # __________________________________________________________________________________________________________________
    def get_return_period(self, height_of_rainfall, duration):
        """
        calculate the return period, when the height of rainfall and the duration are given

        Args:
            height_of_rainfall (float): in [mm]
            duration (int | float | list | numpy.ndarray | pandas.Series): in minutes

        Returns:
            int | float | list | numpy.ndarray | pandas.Series: return period in years
        """
        u, w = self.parameters.get_u_w(duration)
        return np.exp((height_of_rainfall - u) / w)

    # __________________________________________________________________________________________________________________
    def get_duration(self, height_of_rainfall, return_period):
        """
        calculate the duration, when the height of rainfall and the return period are given

        Args:
            height_of_rainfall (float): in [mm]
            return_period (float): in years

        Returns:
            float: duration in minutes
        """
        return newton(lambda d: self.depth_of_rainfall(d, return_period) - height_of_rainfall, x0=1)

    # __________________________________________________________________________________________________________________
    def result_table(self, durations=None, return_periods=None, add_names=False):
        """
        get a standard idf table of rainfall depth with return periods as columns and durations as rows

        Args:
            durations (list | numpy.ndarray): list of durations in minutes for the table
            return_periods (list): list of return periods in years for the table
            add_names (bool): weather to use expressive names as index-&column-label

        Returns:
            pandas.DataFrame: idf table
        """
        if durations is None:
            durations = self.duration_steps

        if return_periods is None:
            return_periods = [1, 2, 3, 5, 10, 20, 25, 30, 50, 75, 100]

        result_table = dict()
        for t in return_periods:
            result_table[t] = self.depth_of_rainfall(durations, t)

        result_table = pd.DataFrame(result_table, index=durations)

        if add_names:
            result_table.index.name = 'duration (min)'
            result_table.columns = pd.MultiIndex.from_tuples([(rp, round(1 / rp, 3)) for rp in result_table.columns])
            result_table.columns.names = ['return period (a)', 'frequency (1/a)']
        return result_table

    ####################################################################################################################
    def result_figure(self, min_duration=5.0, max_duration=720.0, logx=False, return_periods=None, color=True, ax=None,
                      linestyle=None, add_interim=False):
        duration_steps = np.arange(min_duration, max_duration + 1, 1)

        if return_periods is None:
            return_periods = [1, 2, 5, 10, 50, 100]

        table = self.result_table(durations=duration_steps, return_periods=return_periods)
        if color:
            table.columns.name = 'T$\\mathsf{_N}$ in a'
        ax = table.plot(color=(None if color else 'black'), logx=logx, legend=color, ax=ax, ls=linestyle)

        for _, return_time in enumerate(return_periods):
            if add_interim:
                p = self.parameters.measured_points(return_time, max_duration=max_duration)
                # p = measured_points(self, return_time, max_duration=max_duration)
                ax.plot(p, 'k' + 'x')

            if not color:
                x, y = list(table[return_time].tail(1).items())[0]
                ax.text(x + 10, y, '{} a'.format(return_time), verticalalignment='center', horizontalalignment='left',
                        # bbox=dict(facecolor='white', alpha=1.0, lw=1)
                        )

        ax.tick_params(axis='both', which='both', direction='out')
        ax.set_xlabel('Duration D in min')
        ax.set_ylabel('Rainfall h$\\mathsf{_N}$ in mm')
        ax.set_title('IDF curves')

        fig = ax.get_figure()

        cm_to_inch = 2.54
        fig.set_size_inches(h=21 / cm_to_inch, w=29.7 / cm_to_inch)  # (11.69, 8.27)
        fig.tight_layout()
        return fig, ax

    ####################################################################################################################
    @classmethod
    def command_line_tool(cls):
        user = heavy_rain_parser()

        # --------------------------------------------------
        # use the same directory as the input file and make as subdir with the name of the input_file + "_idf_data"
        out = '{label}_idf_data'.format(label='.'.join(user.input.split('.')[:-1]))

        if not path.isdir(out):
            mkdir(out)
            action = 'Creating'
        else:
            action = 'Using'

        print('{} the subfolder "{}" for the interim- and final-results.'.format(action, out))

        fn_pattern = path.join(out, 'idf_{}')

        # --------------------------------------------------
        idf = cls(series_kind=user.series_kind, worksheet=user.worksheet, extended_durations=True)

        # --------------------------------------------------
        parameters_fn = fn_pattern.format('parameters.yaml')

        if path.isfile(parameters_fn):
            print('Found existing interim-results in "{}" and using them for calculations.'.format(parameters_fn))
        else:
            print('Start reading the time-series {} for the analysis.'.format(user.input))
            ts = import_series(user.input).replace(0, np.NaN).dropna()
            # --------------------------------------------------
            idf.set_series(ts)
            print('Finished reading.')

        # --------------------------------------------------
        idf.auto_save_parameters(parameters_fn)

        # --------------------------------------------------
        h = user.height_of_rainfall
        r = user.flow_rate_of_rainfall
        d = user.duration
        t = user.return_period

        if r is not None:
            if h is None and d is not None:
                h = rate2height(rain_flow_rate=r, duration=d)

            elif d is None and h is not None:
                d = h / r * 1000 / 6

        if user.r_720_1:
            d = 720
            t = 1

        if any((h, d, t)):
            if all((d, t)):
                pass

            elif all((d, h)):
                t = idf.get_return_period(h, d)
                print('The return period is {:0.1f} years.'.format(t))

            elif all((h, t)):
                d = idf.get_duration(h, t)
                print('The duration is {:0.1f} minutes.'.format(d))

            print('Resultierende Regenhöhe h_N(T_n={t:0.1f}a, D={d:0.1f}min) = {h:0.2f} mm'
                  ''.format(t=t, d=d, h=idf.depth_of_rainfall(d, t)))
            print('Resultierende Regenspende r_N(T_n={t:0.1f}a, D={d:0.1f}min) = {r:0.2f} L/(s*ha)'
                  ''.format(t=t, d=d, r=idf.rain_flow_rate(d, t)))

        # --------------------------------------------------
        if user.plot:
            fig, ax = idf.result_figure()
            plot_fn = fn_pattern.format('curves_plot.png')
            fig.savefig(plot_fn, dpi=260)
            plt.close(fig)
            show_file(plot_fn)
            print('Created the IDF-curves-plot and saved the file as "{}".'.format(plot_fn))

        # --------------------------------------------------
        if user.export_table:
            table = idf.result_table(add_names=True)
            print(table.round(2).to_string())
            table_fn = fn_pattern.format('table.csv')
            table.to_csv(table_fn, sep=';', decimal=',', float_format='%0.2f')
            print('Created the IDF-curves-plot and saved the file as "{}".'.format(table_fn))

    ####################################################################################################################
    def get_rainfall_sum_frame(self, series=None, durations=None):
        """

        Args:
            series (pandas.Series):
            durations (list): list of durations in minutes which are of interest (default: pre defined
            durations)

        Returns:
            pandas.DataFrame: rain sum depending of the duration per datetimeindex
        """
        if durations is None:
            durations = self.duration_steps

        if series is None:
            if self._rainfall_sum_frame is not None:
                return self._rainfall_sum_frame
            ts = self.series.copy()
            freq = self._freq
        else:
            freq = guess_freq(series.index)
            ts = series.copy()
            ts = ts.asfreq(freq).fillna(0)
            # ts = series.replace(0, np.NaN).dropna()

        df = pd.DataFrame(index=ts.index)
        # df = dict()

        freq_num = delta2min(freq)

        for d in frame_looper(ts.index.size, columns=durations, label='rainfall_sum'):
            if d % freq_num != 0:
                warnings.warn('Using durations (= {} minutes), '
                              'which are not a multiple of the base frequency (= {} minutes) of the series, '
                              'will lead to misinterpretations.'.format(d, freq_num))
            df[d] = ts.rolling(pd.Timedelta(minutes=d)).sum()

        # printable_names (bool): if durations should be as readable in dataframe, else in minutes
        # df = df.rename(minutes_readable, axis=0)

        return df#.round(2)

    @property
    def rainfall_sum_frame(self):
        """
        get the return periods over the whole time-series for the default duration steps.

        Returns:
            pandas.DataFrame: data-frame of return periods where the columns are the duration steps
        """
        if self._rainfall_sum_frame is None:
            self._rainfall_sum_frame = self.get_rainfall_sum_frame()
        return self._rainfall_sum_frame

    ####################################################################################################################
    def get_return_periods_frame(self, series=None, durations=None):
        """

        Args:
            series (pandas.Series):
            durations (list): list of durations in minutes which are of interest (default: pre defined durations)

        Returns:
            pandas.DataFrame: return periods depending of the duration per datetimeindex
        """
        sums = self.get_rainfall_sum_frame(series=series, durations=durations)
        df = pd.DataFrame(index=sums.index)
        # df = dict()
        for d in frame_looper(sums.index.size, columns=sums.columns, label='return_periods'):
            df[d] = self.get_return_period(height_of_rainfall=sums[d][sums[d] >= 0.1], duration=d)
        return df#.fillna(0)#.round(2)

    @property
    def return_periods_frame(self):
        """
        get the return periods over the whole time-series for the default duration steps.

        Returns:
            pandas.DataFrame: data-frame of return periods where the columns are the duration steps
        """
        if self._return_periods_frame is None:
            self._return_periods_frame = self.get_return_periods_frame()
        return self._return_periods_frame

    def write_return_periods_frame(self, filename, **kwargs):
        """save the return-periods dataframe as a parquet-file to save computation time."""
        df = self.return_periods_frame.copy()
        df.columns = df.columns.to_series().astype(str)
        df.round(2).to_parquet(filename, **kwargs)

    def read_return_periods_frame(self, filename, **kwargs):
        """read the return-periods dataframe as a parquet-file to save computation time."""
        df = pd.read_parquet(filename, **kwargs)
        df.columns = df.columns.to_series().astype(int)
        self._return_periods_frame = df

    def auto_save_return_periods_frame(self, filename):
        """auto-save the return-periods dataframe as a parquet-file to save computation time."""
        if path.isfile(filename):
            self.read_return_periods_frame(filename)
        else:
            self.write_return_periods_frame(filename)

    ####################################################################################################################
    @property
    def rain_events(self):
        """
        get the all the rain events of the time-series

        default minimal gap between events is 4 hours

        Returns:
            pandas.DataFrame: data-frame of events with start-, end-time and duration
        """
        if self._rain_events is None:
            events = rain_events(self.series)
            events[COL.DUR] = event_duration(events)
            events[COL.LP] = agg_events(events, self.series, 'sum').round(1)
            events[COL.LAST] = events[COL.START] - events[COL.END].shift()
            # events = events.sort_values(by=COL.LP, ascending=False)
            self._rain_events = events

        return self._rain_events

    def write_rain_events(self, filename, sep=';', decimal='.'):
        """save the rain-events dataframe as a csv-file for external use or to save computation time."""
        self.rain_events.to_csv(filename, index=False, sep=sep, decimal=decimal)

    def read_rain_events(self, filename, sep=';', decimal='.'):
        """read the rain-events dataframe as a csv-file to save computation time."""
        events = pd.read_csv(filename, skipinitialspace=True, sep=sep, decimal=decimal)
        events[COL.START] = pd.to_datetime(events[COL.START])
        events[COL.END] = pd.to_datetime(events[COL.END])
        events[COL.DUR] = pd.to_timedelta(events[COL.DUR])
        events[COL.LAST] = pd.to_timedelta(events[COL.LAST])
        self._rain_events = events

    def auto_save_rain_events(self, filename, sep=';', decimal='.'):
        """auto-save the rain-events dataframe as a csv-file to save computation time."""
        if path.isfile(filename):
            self.read_rain_events(filename, sep=sep, decimal=decimal)
        else:
            self.write_rain_events(filename, sep=sep, decimal=decimal)

    def add_max_return_periods_to_events(self, events):
        if COL.MAX_PERIOD not in events:
            max_periods = self.return_periods_frame.max(axis=1)
            max_periods_duration = self.return_periods_frame.idxmax(axis=1)
            datetime_max = agg_events(events, max_periods, 'idxmax')
            # TODO: where are values NaN ???
            datetime_max = np.where(np.isnan(datetime_max), events[COL.START].values, datetime_max)
            events[COL.MAX_PERIOD] = max_periods[datetime_max].values
            events[COL.MAX_PERIOD_DURATION] = max_periods_duration[datetime_max].values

    ####################################################################################################################
    def event_report(self, filename, min_event_rain_sum=25, min_return_period=0.5, durations=None):
        """
        create pdf file with the biggest rain events
        for each event is represented by a plot of the rain series
        and a IDF analysis where the return periods are calculated

        Args:
            filename (str): path (directory + filename) for the created pdf-report
            min_event_rain_sum (float): only events with a bigger rain sum will be created
            min_return_period (float): only events with a bigger return period will be analysed
                                       (the plot will be created anyway)
            out_path (str): path and filename of the final report
            durations (list[int]): analysed durations
                        (default: [5, 10, 15, 20, 30, 45, 60, 90, 120, 180, 240, 360, 540, 720, 1080, 1440, 2880, 4320])
        """
        if durations is None:
            durations = self.parameters.durations

        events = self.rain_events
        self.add_max_return_periods_to_events(events)

        main_events = events[(events[COL.LP] > min_event_rain_sum) & (events[COL.MAX_PERIOD] > min_return_period)].sort_values(by=COL.MAX_PERIOD, ascending=False).to_dict(
            orient='index')

        unit = 'mm'
        column_name = 'Precipitation'

        pdf = PdfPages(filename)

        for _, event in tqdm(main_events.items()):
            fig, caption = self.event_plot(event, durations=durations, min_return_period=min_return_period,
                                           unit=unit, column_name=column_name)

            # -------------------------------------
            fig.get_axes()[0].set_title(caption + '\n\n\n')

            # DIN A4
            fig.set_size_inches(w=8.27, h=11.69)
            fig.tight_layout()
            pdf.savefig(fig)
            plt.close(fig)

        pdf.close()

    def event_plot(self, event, durations=None, unit='mm', column_name='Precipitation', min_return_period=1):
        if isinstance(event, pd.Series):
            event = event.to_dict()

        plot_range = slice(event[COL.START] - pd.Timedelta(self._freq), event[COL.END] + pd.Timedelta(self._freq))

        if COL.MAX_PERIOD not in event:
            return_periods_frame = self.return_periods_frame[plot_range]
            event[COL.MAX_PERIOD] = return_periods_frame.max().max()
            event[COL.MAX_PERIOD_DURATION] = return_periods_frame.max().idxmax()

        ts = self.series[plot_range].resample(self._freq).sum().fillna(0).copy()

        # -------------------------------------
        fig = plt.figure()

        if event[COL.MAX_PERIOD] < min_return_period:
            rain_ax = fig.add_subplot(111)

        else:
            if durations is not None:
                max_dur = max(durations)
            else:
                max_dur = max(self.duration_steps)

            return_periods_frame_extended = self.get_return_periods_frame(
                self.series[event[COL.START] - pd.Timedelta(minutes=max_dur):
                            event[COL.END] + pd.Timedelta(self._freq)].resample(self._freq).sum(),
                durations=durations
            )

            idf_bar_ax = fig.add_subplot(211)
            idf_bar_ax = idf_bar_axes(idf_bar_ax, return_periods_frame_extended)
            rain_ax = fig.add_subplot(212, sharex=idf_bar_ax)

        # -------------------------------------
        ts_sum, minutes = resample_rain_series(ts)
        rain_ax = rain_bar_plot(ts_sum, rain_ax)
        rain_ax.set_ylabel('{} in {}/{}min'.format(column_name, unit, minutes if minutes != 1 else ''))
        rain_ax.set_xlim(ts.index[0], ts.index[-1])

        return fig, event_caption(event, unit)

    ####################################################################################################################
    def event_return_period_report(self, filename, min_return_period=1):
        events = self.rain_events
        self.add_max_return_periods_to_events(events)

        main_events = events[events[COL.MAX_PERIOD] > min_return_period].sort_values(by=COL.MAX_PERIOD, ascending=False)

        pdf = PdfPages(filename)

        for _, event in tqdm(main_events.to_dict(orient='index').items()):
            fig, ax = self.return_period_event_figure(event)

            # -------------------------------------
            # DIN A4
            fig.set_size_inches(h=11.69 / 2, w=8.27)
            fig.tight_layout()
            pdf.savefig(fig)
            plt.close(fig)

        pdf.close()

    def return_period_event_figure(self, event):
        period_line = self.return_periods_frame[event[COL.START]:event[COL.END]].max()

        period_line[period_line < 0.75] = np.NaN
        period_line = period_line.dropna()

        ax = period_line.plot()  # type: plt.Axes

        ax.set_title('rain event\n'
                     'between {} and {}\n'
                     'with a total sum of {:0.1f} mm\n'
                     'and a duration of {}\n'
                     'The maximum return period was {:0.2f}a\n'
                     'at a duration of {}.'.format(event[COL.START].strftime('%Y-%m-%d %H:%M'),
                                                   event[COL.END].strftime('%Y-%m-%d %H:%M'),
                                                   event[COL.LP],
                                                   event[COL.END] - event[COL.START],
                                                   period_line.max(),
                                                   period_line.idxmax()))

        ax.set_xticks(period_line.index)
        ax.set_xticklabels([minutes_readable(m) for m in period_line.index])
        ax.set_xlabel('duration steps')
        ax.set_ylabel('return period in years')
        return ax.get_figure(), ax



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from .idf_class import IntensityDurationFrequencyAnalyse

"""
heavy rain as a function of the duration and the return period acc. to DWA-A 531 (2012)

This program reads the measurement data of the rainfall from of a file in nieda-format
and calculates the distribution of the rainfall as a function of the return period and the duration

for duration steps up to 12 hours and return period in a range of '0.5 <= T_n <= 100'
"""



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.lines import Line2D
from webbrowser import open as show_file

from . import IntensityDurationFrequencyAnalyse
from .definitions import *
from .little_helpers import minutes_readable
from .sww_utils import agg_events


def return_period_scatter(idf: IntensityDurationFrequencyAnalyse, filename='all_events_max_return_period.pdf', min_return_period=0.5, durations=None):
    if durations is None:
        durations = [5, 10, 15, 20, 30, 45, 60, 90, 120, 180, 240, 360, 540, 720, 1080, 1440, 2880, 4320]

    dur_short = durations[:durations.index(90)]
    dur_long = durations[durations.index(90):]

    events = idf.rain_events
    events[COL.LP] = agg_events(events, idf.series, 'sum')
    events = events[events[COL.LP] > 25].copy()

    tn_long_list = dict()
    tn_short_list = dict()

    for _, event in events.iterrows():
        start = event[COL.START]
        end = event[COL.END]
        # save true
        idf_table = idf.return_periods_frame[start:end]
        idf_table = idf_table.rename(minutes_readable, axis=0)
        # idf_table[idf_table < min_return_period] = np.NaN

        tn = idf_table.loc[start:end]
        tn_short = tn[dur_short].max().max()
        tn_long = tn[dur_long].max().max()

        if tn_long > tn_short:
            tn_long_list[start] = tn_long
        else:
            tn_short_list[start] = tn_short

    print(tn_short_list)
    print(tn_long_list)

    # check()
    fig, ax = plt.subplots()

    ax.scatter(x=list(tn_short_list.keys()), y=list(tn_short_list.values()), color='red')
    ax.scatter(x=list(tn_long_list.keys()), y=list(tn_long_list.values()), color='blue')
    fig = ax.get_figure()

    ax.set_ylabel('Return Period in a')

    def line_in_legend(color=None, marker=None, lw=None, ls=None, **kwargs):
        return Line2D([0], [0], color=color, marker=marker, linewidth=lw, linestyle=ls, **kwargs)

    custom_lines = list()
    custom_lines.append(line_in_legend(color='red', marker='o', lw=0))
    custom_lines.append(line_in_legend(color='blue', marker='o', lw=0))
    # -----------------
    l1 = ax.legend(custom_lines, ['< 60 min', '> 60 min'], loc='best', title='max Duration')
    ax.add_artist(l1)

    # -----------------
    # DIN A4
    fig.set_size_inches(w=7, h=5)
    fig.tight_layout()
    fig.savefig(filename)
    plt.close(fig)


def result_plot_v2(idf: IntensityDurationFrequencyAnalyse, filename, min_duration=5.0, max_duration=720.0, logx=False, show=False):
    duration_steps = np.arange(min_duration, max_duration + 1, 1)
    colors = ['r', 'g', 'b', 'y', 'm']

    # return_periods = [0.5, 1, 10, 50, 100]
    return_periods = [1, 2, 5, 10, 50]

    table = idf.result_table(durations=duration_steps, return_periods=return_periods)
    table.index = pd.to_timedelta(table.index, unit='m')
    ax = table.plot(color=colors, logx=logx)

    ax.tick_params(axis='both', which='both', direction='out')

    for i in range(len(return_periods)):
        return_time = return_periods[i]
        color = colors[i]
        p = idf.parameters.measured_points(return_time, max_duration=max_duration)
        p.index = pd.to_timedelta(p.index, unit='m')
        ax.plot(p, color + 'x')

        # plt.text(max_duration * ((10 - offset) / 10), depth_of_rainfall(max_duration * ((10 - offset) / 10),
        #                                                                 return_time, parameter_1,
        #                                                                 parameter_2) + offset, '$T_n=$' + str(
        #                                                                 return_time))

    ax.set_xlabel('Dauerstufe $D$ in $[min]$')
    ax.set_ylabel('Regenhöhe $h_N$ in $[mm]$')
    ax.set_title('Regenhöhenlinien')
    ax.legend(title='$T_n$= ... [a]')

    if max_duration > 1.5 * 60:
        pass
    else:
        pass

    major_ticks = np.array([d * 60 * 1.0e9 for d in idf.duration_steps if d <= max_duration])
    # minor_ticks = pd.date_range("00:00", "23:59", freq='15T').time
    # print(major_ticks)
    # exit()
    ax.set_xticks(major_ticks)
    # print(ax.get_xticks())
    from matplotlib import ticker

    def time_ticks(x, _):
        x = pd.to_timedelta(x, unit='ns').total_seconds() / 60
        h = int(x / 60)
        m = int(x % 60)
        s = ''
        if h:
            s += '{}h'.format(h)
        if m:
            s += '{}min'.format(m)
        return s

    formatter = ticker.FuncFormatter(time_ticks)
    ax.xaxis.set_major_formatter(formatter)
    # print(ax.get_xticks())
    # plt.axis([0, max_duration, 0, depth_of_rainfall(max_duration,
    #                                                 return_periods[len(return_periods) - 1],
    #                                                 parameter_1, parameter_2) + 10])

    fig = ax.get_figure()

    cm_to_inch = 2.54
    fig.set_size_inches(h=21 / cm_to_inch, w=29.7 / cm_to_inch)  # (11.69, 8.27)
    fig.tight_layout()
    fig.savefig(filename, dpi=260)
    plt.close(fig)
    if show:
        show_file(filename)
    return filename



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import argparse

from .definitions import SERIES, METHOD


class Borders(object):
    def __init__(self, min_=None, max_=None, unit=''):
        self.min_ = min_
        self.max_ = max_
        self.unit = unit

    def __str__(self):
        s = ''
        if self.min_ is not None:
            s += '>= {} {unit}'.format(self.min_, unit=self.unit)
        if self.min_ is not None and self.max_ is not None:
            s += ' and '
        if self.max_ is not None:
            s += '<= {} {unit}'.format(self.max_, unit=self.unit)
        return s

    def __contains__(self, item):
        b = True
        if self.min_ is not None:
            b &= item >= self.min_
        if self.max_ is not None:
            b &= item <= self.max_
        return b

    def __iter__(self):
        return iter([str(self)])


def heavy_rain_parser():
    calc_help = ' (If two of the three variables ' \
                '(rainfall (height or flow-rate), duration, return period) are given, ' \
                'the third variable is calculated.)'
    parser = argparse.ArgumentParser()
    parser.description = 'heavy rain as a function of the duration and the return period acc. to DWA-A 531 (2012)\n' \
                         'All files will be saved in the same directory of the input file ' \
                         'but in a subfolder called like the inputfile + "_idf_data". ' \
                         'Inside this folder a file called "idf_parameter.yaml"-file will be saved and ' \
                         'contains interim-calculation-results and will be automatically reloaded on the next call.'

    parser.add_argument('-i', '--input',
                        help='input file with the rain time-series (csv or parquet)',
                        required=True)
    # --------------------------------------------
    parser.add_argument('-ws', '--worksheet',
                        help='From which worksheet the recommendations for calculating the parameters should be taken.',
                        default=METHOD.KOSTRA,
                        required=False, type=str, choices=METHOD.OPTIONS)
    parser.add_argument('-kind', '--series_kind',
                        help='The kind of series used for the calculation. '
                             '(Calculation with partial series is more precise and recommended.)',
                        default=SERIES.PARTIAL,
                        required=False, type=str, choices=SERIES.OPTIONS)
    # --------------------------------------------
    parser.add_argument('-t', '--return_period',
                        help='return period in years' + calc_help,
                        required=False, type=float, choices=Borders(0.5, 100, 'a'))
    parser.add_argument('-d', '--duration',
                        help='duration in minutes' + calc_help,
                        required=False, type=float, choices=Borders(5, 6 * 24 * 60, 'min'))
    parser.add_argument('-r', '--flow_rate_of_rainfall',
                        help='rainfall in Liter/(s * ha)' + calc_help,
                        required=False, type=float, choices=Borders(0, unit='L/(s*ha)'))
    parser.add_argument('-h_N', '--height_of_rainfall',
                        help='rainfall in mm or Liter/m^2' + calc_help,
                        required=False, type=float, choices=Borders(0, unit='mm'))
    # --------------------------------------------
    parser.add_argument('--r_720_1',
                        help='design rainfall with a duration of 720 minutes (=12 h) and a return period of 1 year',
                        required=False, action='store_true')
    # --------------------------------------------
    parser.add_argument('--plot',
                        help='get a plot of the idf relationship',
                        required=False, action='store_true')
    # --------------------------------------------
    parser.add_argument('--export_table',
                        help='get a table of the most frequent used values',
                        required=False, action='store_true')
    # --------------------------------------------
    return parser.parse_args()



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from . import IntensityDurationFrequencyAnalyse

IntensityDurationFrequencyAnalyse.command_line_tool()



import warnings

from math import floor, e

import numpy as np
import pandas as pd
from tqdm import tqdm

from scipy.stats import linregress

from .definitions import PARAM, SERIES, COL
from .sww_utils import year_delta, guess_freq, rain_events, agg_events


def annual_series(rolling_sum_values, year_index):
    """
    create an annual series of the maximum overlapping sum per year and calculate the "u" and "w" parameters
    acc. to DWA-A 531 chap. 5.1.5

    Args:
        rolling_sum_values (numpy.ndarray): array with maximum rolling sum per event per year.
        year_index (numpy.ndarray): array with year of the event.

    Returns:
        tuple[float, float]: parameter u and w from the annual series for a specific duration step as a tuple
    """
    annually_series = pd.Series(rolling_sum_values).groupby(year_index).max().values
    # annually_series = pd.Series(data=rolling_sum_values,
    #                             index=events[COL.START].values).resample('AS').max().index
    annually_series = np.sort(annually_series)[::-1]

    sample_size = annually_series.size
    index = np.arange(sample_size) + 1
    x = -np.log(np.log((sample_size + 0.2) / (sample_size - index + 0.6)))

    return _lin_regress(x, annually_series)


def _plotting_formula(k, l, m):
    """
    plotting function acc. to DWA-A 531 chap. 5.1.3 for the partial series

    Args:
        k (float): running index
        l (float): sample size
        m (float): measurement period

    Returns:
        float: estimated empirical return period
    """
    return (l + 0.2) * m / ((k - 0.4) * l)


def partial_series(rolling_sum_values, measurement_period):
    """
    create an partial series of the largest overlapping sums and calculate the "u" and "w" parameters
    acc. to DWA-A 531 chap. 5.1.4

    Args:
        rolling_sum_values (numpy.ndarray): array with maximum rolling sum per event
        measurement_period (float): in years

    Returns:
        tuple[float, float]: parameter u and w from the partial series for a specific duration step as a tuple
    """
    partially_series = rolling_sum_values
    partially_series = np.sort(partially_series)[::-1]

    # use only the (2-3 multiplied with the number of measuring years) of the biggest
    # values in the database (-> acc. to ATV-A 121 chap. 4.3; DWA-A 531 chap. 4.4)
    # as an requirement for the extreme value distribution
    threshold_sample_size = int(floor(measurement_period * e))
    partially_series = partially_series[:threshold_sample_size]

    sample_size = threshold_sample_size
    index = np.arange(sample_size) + 1
    log_return_periods = np.log(_plotting_formula(index, sample_size, measurement_period))

    return _lin_regress(log_return_periods, partially_series)


def _lin_regress2(x, y):
    sample_size = len(x)
    x_mean = np.mean(x)
    y_mean = np.mean(y)

    slope = ((x * y).sum() - sample_size * y_mean * x_mean) / \
        ((x ** 2).sum() - sample_size * x_mean ** 2)

    intercept = y_mean - slope * x_mean
    return {PARAM.U: intercept, PARAM.W: slope}


def _lin_regress(x, y):
    res = linregress(x, y)
    return {PARAM.U: res.intercept, PARAM.W: res.slope}


def _improve_factor(interval):
    """
    correction factor acc. to DWA-A 531 chap. 4.3

    Args:
        interval (float): length of the interval: number of observations per duration

    Returns:
        float: correction factor
    """
    improve_factor = {1: 1.14,
                      2: 1.07,
                      3: 1.04,
                      4: 1.03,
                      5: 1.00,
                      6: 1.00}

    return np.interp(interval,
                     list(improve_factor.keys()),
                     list(improve_factor.values()))


def calculate_u_w(file_input, duration_steps, series_kind):
    """
    statistical analysis for each duration step acc. to DWA-A 531 chap. 5.1
    save the parameters of the distribution function as interim results
    acc. to DWA-A 531 chap. 4.4: use the annual series only for measurement periods over 20 years


    Args:
        file_input (pandas.Series): precipitation data
        duration_steps (list[int] | numpy.ndarray): in minutes
        series_kind (str): 'annual' or 'partial'

    Returns:
        dict: with key=durations and values=dict(u, w)
    """
    ts = file_input.copy()
    # -------------------------------
    # measuring time in years
    measurement_start, measurement_end = ts.index[[0, -1]]
    measurement_period = (measurement_end - measurement_start) / year_delta(years=1)
    if round(measurement_period, 1) < 10:
        warnings.warn("The measurement period is too short. The results may be inaccurate! "
                      "It is recommended to use at least ten years. "
                      "(-> Currently {}a used)".format(measurement_period))

    # -------------------------------
    base_frequency = guess_freq(ts.index)  # DateOffset/Timedelta

    # ------------------------------------------------------------------------------------------------------------------
    interim_results = dict()

    # -------------------------------
    # acc. to DWA-A 531 chap. 4.2:
    # The values must be independent of each other for the statistical evaluations.
    # estimated four hours acc. (Schilling, 1984)
    # for larger durations - use the duration as minimal gap
    min_gap_schilling = pd.Timedelta(hours=4)

    # --------------
    # if
    # use only duration for splitting events
    # may increase design-rain-height of smaller durations

    # -------------------------------
    pbar = tqdm(duration_steps, desc='Calculating Parameters u and w')
    for duration_integer in pbar:
        pbar.set_description('Calculating Parameters u and w for duration {:0.0f}'.format(duration_integer))

        duration = pd.Timedelta(minutes=duration_integer)

        if duration < pd.Timedelta(base_frequency):
            continue

        if duration < min_gap_schilling:
            min_gap = min_gap_schilling
        else:
            min_gap = duration

        events = rain_events(ts, min_gap=min_gap)

        # correction factor acc. to DWA-A 531 chap. 4.3
        improve = _improve_factor(duration / base_frequency)

        roll_sum = ts.rolling(duration).sum()

        # events[COL.rolling_sum_valuesAX_OVERLAPPING_SUM] = agg_events(events, roll_sum, 'max') * improve
        rolling_sum_values = agg_events(events, roll_sum, 'max') * improve

        if series_kind == SERIES.ANNUAL:
            interim_results[duration_integer] = annual_series(rolling_sum_values, events[COL.START].dt.year.values)
        elif series_kind == SERIES.PARTIAL:
            interim_results[duration_integer] = partial_series(rolling_sum_values, measurement_period)
        else:
            raise NotImplementedError

    return interim_results



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd
from tqdm import tqdm

from idf_analysis.definitions import COL


def delta2min(time_delta):
    """
    convert timedelta to float in minutes

    Args:
        time_delta (pandas.Timedelta, pandas.DateOffset):

    Returns:
        float: the timedelta in minutes
    """
    if isinstance(time_delta, pd.DateOffset):
        time_delta = time_delta.delta
    return int(time_delta.total_seconds() / 60)


def minutes_readable(minutes):
    """
    convert the duration in minutes to a more readable form

    Args:
        minutes (float | int): duration in minutes

    Returns:
        str: duration as a string
    """
    if minutes <= 60:
        return '{:0.0f} min'.format(minutes)
    elif 60 < minutes < 60 * 24:
        minutes /= 60
        if minutes % 1:
            fmt = '{:0.1f} h'
        else:
            fmt = '{:0.0f} h'
        return fmt.format(minutes)
    elif 60 * 24 <= minutes:
        minutes /= 60 * 24
        if minutes % 1:
            fmt = '{:0.1f} d'
        else:
            fmt = '{:0.0f} d'
        return fmt.format(minutes)
    else:
        return str(minutes)


def duration_steps_readable(durations):
    """
    convert the durations to a more readable form

    Args:
        durations (list[int | float]): in minutes

    Returns:
        list[str]: of the readable duration list
    """
    return [minutes_readable(i) for i in durations]


def height2rate(height_of_rainfall, duration):
    """
    calculate the specific rain flow rate in [l/(s*ha)]
    if 2 array-like parameters are give, a element-wise calculation will be made.
    So the length of the array must be the same.

    Args:
        height_of_rainfall (float | np.ndarray | pd.Series): height_of_rainfall: in [mm]
        duration (float | np.ndarray | pd.Series): in minutes

    Returns:
        float | np.ndarray | pd.Series: specific rain flow rate in [l/(s*ha)]
    """
    return height_of_rainfall / duration * (1000 / 6)


def rate2height(rain_flow_rate, duration):
    """
    convert the rain flow rate to the height of rainfall in [mm]
    if 2 array-like parameters are give, a element-wise calculation will be made.
    So the length of the array must be the same.

    Args:
        rain_flow_rate (float | np.ndarray | pd.Series): in [l/(s*ha)]
        duration (float | np.ndarray | pd.Series): in minutes

    Returns:
        float | np.ndarray | pd.Series: height of rainfall in [mm]
    """
    return rain_flow_rate * duration / (1000 / 6)


def frame_looper(size, columns, label='return periods'):
    if size > 30000:  # if > 3 weeks, use a progressbar
        return tqdm(columns, desc=f'calculating {label} data-frame')
    else:
        return columns


def event_caption(event, unit='mm'):
    caption = 'rain event\n'
    if (COL.START in event) and (COL.END in event):
        caption += f' between {event[COL.START]:%Y-%m-%d %H:%M} and {event[COL.END]:%Y-%m-%d %H:%M}'

    if COL.LP in event:
        caption += f' with a total sum of {event[COL.LP]:0.1f} {unit}'

    if COL.DUR in event:
        caption += f' and a duration of {event[COL.DUR]}'

    caption += '.\n'

    if COL.MAX_PERIOD in event:
        caption += f' The maximum return period was {return_period_formatter(event[COL.MAX_PERIOD])} a\n'

        if COL.MAX_PERIOD_DURATION in event:
            caption += f' at a duration of {minutes_readable(event[COL.MAX_PERIOD_DURATION])}.'

    return caption


def return_period_formatter(t):
    if t < 1:
        return '< 1'
    elif t > 200:
        return '$\\gg$ 100'
    elif t > 100:
        return '> 100'
    else:
        return f'{t:0.1f}'



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"


class METHOD:
    # DWA = 'DWA-A_531'  # (2012)
    KOSTRA = 'KOSTRA'
    CONVECTIVE_ADVECTIVE = 'convective_vs_advective'

    ATV = 'ATV-A_121'  # (1985)

    DWA_adv = CONVECTIVE_ADVECTIVE

    # ÖKOSTRA ohne Dauerstufenausgleich nach Hammer (1993) "Eine optimierte Starkniederschalgsauswertung"
    OWUNDA = 'OWUNDA'

    OPTIONS = [KOSTRA, CONVECTIVE_ADVECTIVE, ATV]


class SERIES:
    # base series for calculation
    PARTIAL = 'partial'
    ANNUAL = 'annual'

    OPTIONS = [PARTIAL, ANNUAL]


class APPROACH:
    # parameters approach
    LOG1 = 'single_logarithmic'
    LOG2 = 'double_logarithmic'
    HYP = 'hyperbolic'
    LIN = 'linear'


class PARAM:
    SERIES = 'series_kind'  # Art der Serie
    # METHOD = 'Methodik'

    DUR = 'durations'  # Dauerstufen

    # parameters of the distribution of the series
    # parameters_series
    U = 'u'
    W = 'w'
    U_AND_W = U, W

    PARAMS_SERIES = 'parameters_series'
    PARAMS_FINAL = 'parameters_final'

    A = 'a'
    B = 'b'
    FUNCTION = 'function'

    # parameters of the distribution function of the regression
    # parameters_final

    # parameters for the function
    # @staticmethod
    # def A(p):
    #     return '{}_{}'.format('a', p)
    #
    # @staticmethod
    # def B(p):
    #     return '{}_{}'.format('b', p)
    #
    # @staticmethod
    # def FUNCTION(p):
    #     return 'function_{}'.format(p)

    @staticmethod
    def VALUES(p):
        return '{}_values'.format(p)


class COL:
    # column names for the event table
    START = 'start'
    END = 'end'
    DUR = 'duration'
    LP = 'rain_sum'
    MAX_OVERLAPPING_SUM = 'max_overlapping_sum'
    MAX_PERIOD = 'max_return_period'
    MAX_PERIOD_DURATION = 'max_return_period_duration'
    LAST = 'last_event'



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd
import yaml
from collections import OrderedDict


def import_series(filename, series_label='precipitation', index_label='datetime', csv_reader_args=None):
    """

    Args:
        filename:
        series_label:
        index_label:
        csv_reader_args: for example: sep="," or "." and decimal=";" or ","

    Returns:
        pandas.Series: precipitation series
    """
    if filename.endswith('csv'):
        if csv_reader_args is None:
            csv_reader_args = dict(sep=';', decimal=',')
        try:
            ts = pd.read_csv(filename, index_col=0, header=0, squeeze=True, **csv_reader_args)
            ts.index = pd.to_datetime(ts.index)
            ts.index.name = index_label
            ts.name = series_label
        except:
            raise UserWarning('ERROR | '
                              'Something is wrong with your csv format. The file should only include two columns. '
                              'First column is the date and time index (prefered format is "YYYY-MM-DD HH:MM:SS") '
                              'and second column the precipitation values in mm. '
                              'As a separator use "{sep}" and as decimal sign use "{decimal}".'.format(**csv_reader_args))

        return ts
    elif filename.endswith('parquet'):
        return pd.read_parquet(filename, columns=[series_label])[series_label].rename_axis(index_label, axis='index')
    elif filename.endswith('pkl'):
        return pd.read_pickle(filename).rename(series_label).rename_axis(index_label, axis='index')
    else:
        raise NotImplementedError('Sorry, but only csv, parquet and pickle files are implemented. '
                                  'Maybe there will be more options soon.')

# ------------------------------------------------------------------------------
_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def _dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


yaml.add_representer(OrderedDict, _dict_representer)
yaml.add_constructor(_mapping_tag, _dict_constructor)


def write_yaml(data, fn):
    yaml.dump(data, open(fn, 'w'), default_flow_style=None, width=200)


def read_yaml(filename):
    return yaml.load(open(filename, 'r'), Loader=yaml.FullLoader)



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from collections import OrderedDict

import pandas as pd
import numpy as np
import os

from .definitions import *
from .event_series_analysis import calculate_u_w
from .in_out import write_yaml, read_yaml
from .parameter_formulations import folded_log_formulation, hyperbolic_formulation


class IdfParameters:
    def __init__(self, series_kind=SERIES.PARTIAL, worksheet=METHOD.KOSTRA, extended_durations=False):
        self.series_kind = series_kind

        self._durations = None
        self._set_default_durations(extended_durations)

        self.parameters_series = dict()  # parameters u and w (distribution function) from the event series analysis
        self.parameters_final = dict()  # parameters of the distribution function after the regression

        self.get_approaches(worksheet)

    # -------------------------------------------------------------
    def __bool__(self):
        return bool(self.parameters_series)

    # -------------------------------------------------------------
    def calc_from_series(self, series):
        u_w = calculate_u_w(series, self.durations, self.series_kind)
        for p in PARAM.U_AND_W:
            self.parameters_series[p] = np.array([d[p] for d in u_w.values()])
        self._calc_params()
        self._balance_parameter_change()

    def reverse_engineering(self, idf_table):
        durations = idf_table.index.values
        u = idf_table[1].values
        w_dat = idf_table.sub(u, axis=0)
        # --
        w = list()
        for dur in durations:
            dat = w_dat.loc[dur]
            log_tn_i = np.log(dat.index.values)
            w.append(sum(dat.values * log_tn_i) / sum(log_tn_i ** 2))

        # -------
        # ax = dat.plot(logx=True)
        # fig = ax.get_figure()
        # fig.show()
        # print(results.summary())

        self.durations = durations
        self.parameters_series[PARAM.U] = np.array(u)
        self.parameters_series[PARAM.W] = np.array(w)
        self._calc_params()
        self._balance_parameter_change()

    # -------------------------------------------------------------
    @property
    def durations(self):
        return np.array(self._durations)

    @durations.setter
    def durations(self, durations):
        self._durations = list(durations)

    def _set_default_durations(self, extended_durations=False):
        # Suggestion from ATV-A 121 (ATV-DVWK-Regelwerk 2/2001)

        # sampling points of the duration steps in minutes
        duration_steps = [5, 10, 15, 20, 30, 45, 60, 90]
        # self._duration_steps += [i * 60 for i in [3, 4.5, 6, 7.5, 10, 12, 18]]  # duration steps in hours
        duration_steps += [i * 60 for i in [2, 3, 4, 6, 9, 12, 18]]  # duration steps in hours
        if extended_durations:
            duration_steps += [i * 60 * 24 for i in [1, 2, 3, 4, 5, 6]]  # duration steps in days

        self.durations = duration_steps

    def filter_durations(self, freq_minutes):
        self.limit_duration(lowest=freq_minutes)

    def limit_duration(self, lowest=None, highest=None):
        bool_array = self.durations == False
        if lowest is not None:
            bool_array |= self.durations >= lowest
        if highest is not None:
            bool_array |= self.durations <= highest

        self.durations = self.durations[bool_array]
        for param in PARAM.U_AND_W:
            if param in self.parameters_series:
                self.parameters_series[param] = self.parameters_series[param][bool_array]
        # self.parameters_series[PARAM.W] = self.parameters_series[PARAM.W][bool_array]

    # -------------------------------------------------------------
    def get_approaches(self, worksheet):
        """
        approaches depending on the duration and the parameter

        Args:
            worksheet (str): worksheet name for the analysis:
                - 'DWA-A_531'
                - 'ATV-A_121'
                - 'DWA-A_531_advektiv' (yet not implemented)

        Returns:
            list[dict]: table of approaches depending on the duration and the parameter
        """
        self.parameters_final = read_yaml(os.path.join(os.path.dirname(__file__), 'approaches', worksheet + '.yaml'))

    # def _dur_upper_bound(self, dur_min):
    #     lower_bounds = list(self.parameters_final.keys())
    #     return lower_bounds[lower_bounds.index(dur_min) + 1]

    def _iter_params(self):
        lower_bounds = list(self.parameters_final.keys())
        return zip(lower_bounds, lower_bounds[1:] + [np.inf])

    def _calc_params(self, params_mean=None, duration_mean=None):
        """
        calculate parameters a_u, a_w, b_u and b_w and add it to the dict

        Args:
            params_mean (dict[float]):
            duration_mean (float):

        Returns:
            list[dict]: parameters
        """
        for dur_lower_bound, dur_upper_bound in self._iter_params():
            param_part = (self.durations >= dur_lower_bound) & (self.durations <= dur_upper_bound)

            params_dur = self.parameters_final[dur_lower_bound]

            dur = self.durations[param_part]

            for p in PARAM.U_AND_W:  # u or w
                params = params_dur[p]
                values_series = self.parameters_series[p][param_part]

                approach = params[PARAM.FUNCTION]

                if params_mean:
                    param_mean = params_mean[p]
                else:
                    param_mean = None

                # ----------------------------
                if approach in [APPROACH.LOG1, APPROACH.LOG2]:
                    params[PARAM.A], params[PARAM.B] = folded_log_formulation(dur, values_series, case=approach,
                                                                              param_mean=param_mean,
                                                                              duration_mean=duration_mean)

                # ----------------------------
                elif approach == APPROACH.HYP:
                    a_start = 20.0
                    if PARAM.A in params and not np.isnan(params[PARAM.A]):
                        a_start = params[PARAM.A]

                    b_start = 15.0
                    if PARAM.B in params and not np.isnan(params[PARAM.B]):
                        b_start = params[PARAM.B]

                    params[PARAM.A], params[PARAM.B] = hyperbolic_formulation(dur, values_series, a_start=a_start,
                                                                              b_start=b_start,
                                                                              param_mean=param_mean,
                                                                              duration_mean=duration_mean)

                # ----------------------------
                elif approach == APPROACH.LIN:
                    pass

                # ----------------------------
                else:
                    raise NotImplementedError

                # ----------------------------

    def _balance_parameter_change(self):
        # the balance between the different duration ranges acc. to DWA-A 531 chap. 5.2.4
        duration_step = list(self.parameters_final.keys())[1]
        durations = np.array([duration_step - 0.001, duration_step + 0.001])

        # if the interim results end here

        if any(durations < self.durations.min()) or \
                any(durations > self.durations.max()):
            return

        u, w = self.get_u_w(durations)
        self._calc_params(params_mean=dict(u=np.mean(u), w=np.mean(w)), duration_mean=duration_step)

    def measured_points(self, return_periods, max_duration=None):
        """
        get the calculation results of the rainfall with u and w without the estimation of the formulation

        Args:
            return_periods (float | np.array | list | pd.Series): return period in [a]
            max_duration (float): max duration in [min]

        Returns:
            pd.Series: series with duration as index and the height of the rainfall as data
        """
        interim_results = pd.DataFrame(index=self.durations)
        interim_results.index.name = PARAM.DUR
        interim_results[PARAM.U] = self.parameters_series[PARAM.U]
        interim_results[PARAM.W] = self.parameters_series[PARAM.W]

        if max_duration is not None:
            interim_results = interim_results.loc[:max_duration].copy()

        return pd.Series(index=interim_results.index,
                         data=interim_results['u'] + interim_results['w'] * np.log(return_periods))

    def get_duration_section(self, duration, param):
        for lower, upper in self._iter_params():
            if lower <= duration <= upper:
                return self.parameters_final[lower][param]

    def get_scalar_param(self, p, duration):
        """

        Args:
            p (str): name of the parameter 'u' or 'w'
            duration (float | int): in minutes

        Returns:
            (float, float): u, w
        """
        param = self.get_duration_section(duration, p)

        if param is None:
            return np.NaN

        elif param[PARAM.FUNCTION] == APPROACH.LOG1:
            a = param[PARAM.A]
            b = param[PARAM.B]
            return a + b * np.log(duration)

        elif param[PARAM.FUNCTION] == APPROACH.LOG2:
            a = param[PARAM.A]
            b = param[PARAM.B]
            return np.exp(a) * np.power(duration, b)

        elif param[PARAM.FUNCTION] == APPROACH.HYP:
            a = param[PARAM.A]
            b = param[PARAM.B]
            return a * duration / (duration + b)

        elif param[PARAM.FUNCTION] == APPROACH.LIN:
            return np.interp(duration, self.durations, self.parameters_series[p])

    def get_array_param(self, p, duration):
        """

        Args:
            p (str): name of the parameter 'u' or 'w'
            duration (numpy.ndarray): in minutes

        Returns:
            (numpy.ndarray, numpy.ndarray): u, w
        """
        return np.vectorize(lambda d: self.get_scalar_param(p, d))(duration)

    def get_u_w(self, duration):
        """
        calculate the u and w parameters depending on the durations

        Args:
            duration (numpy.ndarray| list | float | int): in minutes

        Returns:
            (float, float): u and w
        """
        if isinstance(duration, (list, np.ndarray)):
            func = self.get_array_param
        else:
            func = self.get_scalar_param

        return (func(p, duration) for p in PARAM.U_AND_W)

    @classmethod
    def from_interim_results_file(cls, interim_results_fn, worksheet=METHOD.KOSTRA):
        """DEPRECIATED: for compatibility reasons. To use the old file and convert it to the new parameters file"""
        interim_results = pd.read_csv(interim_results_fn, index_col=0, header=0)
        p = cls(worksheet, interim_results)
        return p

    def to_yaml(self, filename):
        def to_basic(a):
            return list([round(float(i), 4) for i in a])
        data = OrderedDict({
            PARAM.SERIES: self.series_kind,
            PARAM.DUR: to_basic(self.durations),
            PARAM.PARAMS_SERIES: {p: to_basic(l) for p, l in self.parameters_series.items()},
            PARAM.PARAMS_FINAL: self.parameters_final
        })
        write_yaml(data, filename)

    @classmethod
    def from_yaml(cls, filename):
        data = read_yaml(filename)
        if isinstance(data, list):
            return cls.from_yaml_depreciated(filename)
        p = cls(series_kind=data[PARAM.SERIES])
        p.durations = data[PARAM.DUR]
        # list to numpy.array
        p.parameters_series = {p: np.array(l) for p, l in data[PARAM.PARAMS_SERIES].items()}
        p.parameters_final = data[PARAM.PARAMS_FINAL]
        return p

    @classmethod
    def from_yaml_depreciated(cls, filename, series_kind=SERIES.PARTIAL):
        data = read_yaml(filename)
        p = cls(series_kind=series_kind)
        p.durations = list()
        p.parameters_series = {PARAM.U: list(), PARAM.W: list()}
        p.parameters_final = dict()
        for part in data:
            start_dur = float(part['von'])
            start_idx = 0
            if start_dur in part[COL.DUR]:
                start_idx = 1
            p._durations += part[COL.DUR][start_idx:]
            part_params = dict()

            for u_or_w in PARAM.U_AND_W:
                part_params[u_or_w] = {PARAM.FUNCTION: part[u_or_w]}
                if part[u_or_w] != APPROACH.LIN:
                    part_params[u_or_w][PARAM.A] = part[f'{PARAM.A}_{u_or_w}']
                    part_params[u_or_w][PARAM.B] = part[f'{PARAM.B}_{u_or_w}']

                p.parameters_series[u_or_w] += part[PARAM.VALUES(u_or_w)][start_idx:]

            p.parameters_final[start_dur] = part_params

        p.parameters_series = {param: np.array(l) for param, l in p.parameters_series.items()}
        if '.yaml' in filename:
            os.rename(filename, filename.replace('.yaml', '_OLD.yaml'))
            p.to_yaml(filename)
        return p



__author__ = "Markus Pichler, Marlies Hierzer"
__credits__ = ["Markus Pichler", "Marlies Hierzer"]
__maintainer__ = "Markus Pichler, Marlies Hierzer"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"


import matplotlib.pyplot as plt
import numpy as np
import pandas
import pandas as pd
from matplotlib.ticker import NullFormatter

from idf_analysis import IntensityDurationFrequencyAnalyse
from idf_analysis.definitions import COL
from idf_analysis.little_helpers import duration_steps_readable, minutes_readable, frame_looper, event_caption
from idf_analysis.sww_utils import (guess_freq, rain_events, event_duration, resample_rain_series, rain_bar_plot,
                                    agg_events, )

COL.MAX_SRI = 'max_SRI_{}'
COL.MAX_SRI_DURATION = 'max_SRI_duration_{}'


####################################################################################################################
def grisa_factor(tn):
    """
    calculates the grisa-factor according to Grisa's formula

    Args:
        tn (float): in [years]

    Returns:
        float: factor
    """
    return 1 + (np.log(tn) / np.log(2))


def next_bigger(v, l):
    return l[next(x for x, val in enumerate(l) if val >= v)]


class SCHMITT:
    # Zuweisung nach Schmitt des SRI über der Wiederkehrperiode
    SRI_TN = {
        1: 1,
        2: 1,
        3: 2,
        5: 2,
        10: 3,
        20: 4,
        25: 4,
        30: 5,
        50: 6,
        75: 6,
        100: 7
    }

    # Erhöhungsfaktoren nach Schmitt für SRI 8,9,10,11,12 basierend auf SRI 7
    # untere und obere Grenze
    MULTI_FACTOR = {
        8: (1.2, 1.39),
        9: (1.4, 1.59),
        10: (1.6, 2.19),
        11: (2.2, 2.78),
        12: (2.8, 2.8),
    }

    VERBAL = {
        (1, 2): 'Starkregen',
        (3, 5): 'intensiver Starkregen',
        (6, 7): 'außergewöhnlicher Starkregen',
        (8, 12): 'extremer Starkregen'
    }

    INDICES_COLOR = {1: (0.69, 0.9, 0.1),
                     2: (0.8, 1, 0.6),
                     3: (0.9, 1, 0.3),
                     4: (1, 0.96, 0),
                     5: (1, 0.63, 0),
                     6: (1, 0.34, 0),
                     7: (1, 0.16, 0),
                     8: (0.97, 0.12, 0.24),
                     9: (1, 0.10, 0.39),
                     10: (0.97, 0.03, 0.51),
                     11: (0.92, 0.08, 0.75),
                     12: (0.66, 0.11, 0.86)}

    INDICES_COLOR_RGB = {1: (176, 230, 25),
                         2: (204, 255, 153),
                         3: (230, 255, 77),
                         4: (255, 244, 0),
                         5: (255, 160, 0),
                         6: (255, 86, 0),
                         7: (255, 40, 0),
                         8: (247, 30, 61),
                         9: (255, 26, 99),
                         10: (247, 9, 130),
                         11: (235, 21, 191),
                         12: (189, 28, 220)}

    INDICES_COLOR_HEX = {1: "#b0e619",
                         2: "#ccff99",
                         3: "#e6ff4d",
                         4: "#fff400",
                         5: "#ffa000",
                         6: "#ff5600",
                         7: "#ff2800",
                         8: "#f71e3d",
                         9: "#ff1a63",
                         10: "#f70982",
                         11: "#eb15bf",
                         12: "#bd1cdc"}


krueger_pfister_verbal = {
    (1, 4): 'moderat',
    (5, 7): 'stark',
    (8, 10): 'heftig',
    (11, 12): 'extrem'
}

grisa_verbal = {
    (1, 2): 'Minor',
    (3, 4): 'Moderate',
    (5, 6): 'Major',
    (7, 8): 'Extreme',
    (9, 10): 'Catastrophic'
}


def cat_dict(cat):
    res = {}
    for num_range, verbal in cat.items():
        for i in range(num_range[0], num_range[1]+1):
            res[i] = verbal
    return res


####################################################################################################################
class HeavyRainfallIndexAnalyse(IntensityDurationFrequencyAnalyse):
    indices = list(range(1, 13))

    class METHODS:
        SCHMITT = 'Schmitt'
        KRUEGER_PFISTER = 'KruegerPfister'
        MUDERSBACH = 'Mudersbach'

        @classmethod
        def all(cls):
            return cls.SCHMITT, cls.KRUEGER_PFISTER, cls.MUDERSBACH

    indices_color = SCHMITT.INDICES_COLOR

    def __init__(self, *args, method=METHODS.SCHMITT, **kwargs):
        IntensityDurationFrequencyAnalyse.__init__(self, *args, **kwargs)
        self.method = method
        self._sri_frame = None

    def set_series(self, series):
        IntensityDurationFrequencyAnalyse.set_series(self, series)
        self._sri_frame = None

    def get_sri(self, height_of_rainfall, duration):
        """
        calculate the heavy rain index (StarkRegenIndex), when the height of rainfall and the duration are given

        Args:
            height_of_rainfall (float): in [mm]
            duration (int | float | list | numpy.ndarray | pandas.Series): in minutes

        Returns:
            int | float | list | numpy.ndarray | pandas.Series: heavy rain index
        """
        tn = self.get_return_period(height_of_rainfall, duration)

        if self.method == self.METHODS.MUDERSBACH:
            if isinstance(tn, (pd.Series, np.ndarray)):
                sri = np.round(1.5 * np.log(tn) + 0.4 * np.log(duration), 0)
                sri[tn <= 1] = 1
                sri[tn >= 100] = 12
                return sri
            else:
                if tn <= 1:
                    return 1
                elif tn >= 100:
                    return 12
                else:
                    return np.round(1.5 * np.log(tn) + 0.4 * np.log(duration), 0)

        elif self.method == self.METHODS.SCHMITT:
            if isinstance(tn, (pd.Series, np.ndarray)):
                breaks = [-np.inf] + list(SCHMITT.SRI_TN.keys()) + [np.inf]
                d = dict(zip(range(11), SCHMITT.SRI_TN.values()))
                sri = pd.cut(tn, breaks, labels=False).replace(d)

                over_100 = tn > 100
                hn_100 = self.depth_of_rainfall(duration, 100)
                breaks2 = [1] + [f[0] for f in SCHMITT.MULTI_FACTOR.values()][1:] + [np.inf]
                d2 = dict(zip(range(len(breaks2) - 1), range(8, 13)))
                sri.loc[over_100] = pd.cut(height_of_rainfall.loc[over_100] / hn_100, breaks2, labels=False).replace(d2)

            else:
                if tn >= 100:
                    hn_100 = self.depth_of_rainfall(duration, 100)
                    for sri, mul in SCHMITT.MULTI_FACTOR.items():
                        if height_of_rainfall <= hn_100 * mul[0]:
                            break
                else:
                    sri = SCHMITT.SRI_TN[next_bigger(tn, list(SCHMITT.SRI_TN.keys()))]

        elif self.method == self.METHODS.KRUEGER_PFISTER:
            h_24h = self.depth_of_rainfall(duration=24 * 60, return_period=tn)
            hn_100 = self.depth_of_rainfall(duration=duration, return_period=100)
            duration_adjustment_factor = height_of_rainfall / h_24h
            intensity_adjustment_factor = height_of_rainfall / hn_100
            sri = grisa_factor(tn) * duration_adjustment_factor * intensity_adjustment_factor
            if isinstance(sri, (pd.Series, np.ndarray)):
                sri[tn < 0.5] = 0
            else:
                if tn < 0.5:
                    return 0
            return np.clip(np.ceil(sri), 0, 12)

        else:
            raise NotImplementedError(f'Method {self.method} not implemented!')
        return sri

    # __________________________________________________________________________________________________________________
    def result_sri_table(self, durations=None):
        """
        get a standard idf table of rainfall depth with return periods as columns and durations as rows

        Args:
            durations (list | numpy.ndarray | None): list of durations in minutes for the table

        Returns:
            pandas.DataFrame: idf table
        """
        idf_table = self.result_table(durations)

        if self.method == self.METHODS.SCHMITT:
            sri_table = idf_table.rename(columns=SCHMITT.SRI_TN)

            for sri, mul in SCHMITT.MULTI_FACTOR.items():
                sri_table[sri] = mul[1] * sri_table[7]

            sri_table = sri_table.loc[:, ~sri_table.columns.duplicated('last')]

        elif self.method == self.METHODS.MUDERSBACH:
            # zuerst eine Tabelle mit den Wiederkehrperioden
            rp_table = pd.DataFrame(index=idf_table.index, columns=range(1, 13))

            # abhängigkeit nach dauerstufe
            a = np.log(rp_table.index.values) * 0.4
            for sri in rp_table.columns:
                rp_table[sri] = np.exp((sri + 0.5 - a) / 1.5)

            rp_table.loc[:, 1] = 1
            # rp_table.loc[:, 12] = 100

            # dann mittels Dauerstufe und Wiederkehrperiode die Regenhöhe errechnen
            sri_table = rp_table.round(1).copy()
            for dur in rp_table.index:
                sri_table.loc[dur] = self.depth_of_rainfall(dur, rp_table.loc[dur])

            # extrapolation vermutlich nicht sehr seriös
            sri_table[rp_table >= 100] = np.NaN
            # sri_table.loc[:12] = self.depth_of_rainfall(sri_table.index.values, 100)
            sri_table[rp_table < 1] = np.NaN
            sri_table = sri_table.astype(float).round(2)
            sri_table = sri_table.fillna(method='ffill', axis=1, limit=None)

        elif self.method == self.METHODS.KRUEGER_PFISTER:
            # duration_adjustment_factor = idf_table.div(idf_table.loc[24 * 60])
            # intensity_adjustment_factor = idf_table.div(idf_table[100].values, axis=0)
            # sri_table = grisa_factor(
            #     idf_table.columns.values) * duration_adjustment_factor * intensity_adjustment_factor
            # sri_table = sri_table.round().astype(int).clip(0,12)

            sri_table = pd.DataFrame(index=idf_table.index)
            sri_vector = (idf_table.loc[1440, 100] * idf_table.loc[:, 100]) / (1 + (np.log(100) / np.log(2)))
            for i in self.indices:
                sri_table[i] = np.sqrt(i * sri_vector)

        else:
            raise NotImplementedError(f'Method or "{self.method}" not implemented! '
                                      f'Please contact the developer for the request to implement it.')

        sri_table.index.name = 'duration in min'
        sri_table.columns.name = 'SRI'
        return sri_table

    def interim_sri_table(self, durations=None):
        """M
        get a table of SRI with return periods as columns and durations as rows

        Args:
            durations (list | numpy.ndarray): list of durations in minutes for the table
            return_periods (list): list of return periods in years for the table

        Returns:
            pandas.DataFrame: idf table
        """

        idf_table = self.result_table(durations)
        sri_table = pd.DataFrame(index=idf_table.index, columns=idf_table.columns)

        if self.method == self.METHODS.SCHMITT:
            for col in sri_table:
                sri_table[col] = SCHMITT.SRI_TN[col]

        elif self.method == self.METHODS.MUDERSBACH:
            sri_table[1] = 1

            a = np.log(sri_table.index.values) * 0.4
            for tn in [2, 3, 5, 10, 20, 25, 30, 50, 75, 100]:
                sri_table[tn] = a + np.log(tn) * 1.5
            sri_table = sri_table.round().astype(int)

        elif self.method == self.METHODS.KRUEGER_PFISTER:
            duration_adjustment_factor = idf_table.div(idf_table.loc[24 * 60])
            intensity_adjustment_factor = idf_table.div(idf_table[100].values, axis=0)
            sri_table = grisa_factor(idf_table.columns.values) * duration_adjustment_factor * intensity_adjustment_factor
            sri_table = sri_table.round().astype(int).clip(0,12)

        else:
            raise NotImplementedError(f'Method {self.method} not implemented!')

        sri_table.index.name = 'duration in min'
        sri_table.columns.name = 'Return Period in a'
        return sri_table

    ####################################################################################################################
    def result_sri_figure(self, duration_steps=None, ax=None, grid=True):
        """
        SRI curves are generated depending on the selected procedure for SRI generation.

        Args:
            duration_steps (list | numpy.ndarray): list of durations in minutes for the table
            ax (plt.Axes): if plot is a subplot give the axes
            grid (bool): if to make a grid

        Returns:
            (matplotlib.pyplot.Figure, matplotlib.pyplot.Axes): figure and axes of the plot
        """
        sri_table = self.result_sri_table(durations=duration_steps)

        ax = sri_table.plot(color=self.indices_color, logx=True, legend=True, ax=ax)

        ax.tick_params(axis='both', which='both', direction='out')
        ax.set_xlabel('Duration D')
        ax.set_ylabel('Rainfall h$\\mathsf{_N}$ in mm')
        # ax.set_xlabel('Dauerstufe D')
        # ax.set_ylabel('Regenhöhe h$\\mathsf{_N}$ in mm')
        # ax.set_title('Starkregenindex - Kurven', fontweight='bold')

        ax.set_xticks([], minor=True)
        ax.set_xticks(sri_table.index)
        ax.set_xlim(*sri_table.index.values[[0, -1]])

        ax.set_xticklabels(duration_steps_readable(sri_table.index))

        ax.set_facecolor('w')
        if grid:
            ax.grid(color='grey', linestyle='-', linewidth=0.3)

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles[::-1], labels[::-1], bbox_to_anchor=(1.02, 1.), loc='upper left',
                  borderaxespad=0., title='SRI')

        fig = ax.get_figure()

        # cm_to_inch = 2.54
        # fig.set_size_inches(h=11 / cm_to_inch, w=25 / cm_to_inch)  # (11.69, 8.27)
        fig.tight_layout()
        return fig, ax

    ####################################################################################################################
    @property
    def sri_frame(self):
        """
        get the return periods over the whole time-series for the default duration steps.

        Returns:
            pandas.DataFrame: data-frame of return periods where the columns are the duration steps
        """
        if self._sri_frame is None:
            self._sri_frame = self.get_sri_frame()
        return self._sri_frame

    def get_sri_frame(self, series=None, durations=None):
        """

        Args:
            series (pandas.Series): precipitation time-series of the time range of interest i.e. of an event
            durations (list): list of durations in minutes which are of interest (default: pre defined durations)

        Returns:
            pandas.DataFrame: index=date-time-index; columns=durations; values=SRI
        """
        # TODO: Probleme bei geringen Regenhöhen, Formel nicht dafür gemacht!!
        sums = self.get_rainfall_sum_frame(series=series, durations=durations)
        df = pd.DataFrame(index=sums.index)
        for d in frame_looper(sums.index.size, columns=sums.columns, label='sri'):
            df[d] = self.get_sri(height_of_rainfall=sums[d][sums[d] > 0.1], duration=d)
        return df#.round(1)

    def add_max_sri_to_events(self, events, series=None):
        """M
        maximum SRI is added to the table

        Args:
            events (list): list of rainfall events

        Returns:
            pandas.DataFrame: table
        """
        if COL.MAX_SRI.format(self.method) not in events:
            events[COL.MAX_SRI.format(self.method)] = None
            events[COL.MAX_SRI_DURATION.format(self.method)] = None

            rainfall_sum_frame = self.get_rainfall_sum_frame(series=series)

            for event_no, event in events.iterrows():
                s = self.get_event_sri_max(event[COL.START], event[COL.END], rainfall_sum_frame=rainfall_sum_frame)
                events.loc[event_no, COL.MAX_SRI.format(self.method)] = s.max()
                events.loc[event_no, COL.MAX_SRI_DURATION.format(self.method)] = s.idxmax()

    def get_event_sri_max(self, start, end, rainfall_sum_frame=None):
        """M
        maximum SRI is added to the table

        Args:
            events (list): list of rainfall events

        Returns:
            pandas.DataFrame: table
        """
        if rainfall_sum_frame is None:
            d = self.rainfall_sum_frame[start:end].max().to_dict()
        else:
            d = rainfall_sum_frame[start:end].max().to_dict()
        sri = dict()
        for dur, h in d.items():
            sri[dur] = self.get_sri(h, dur)
        return pd.Series(sri, name=self.method)

    def sri_bar_axes(self, ax, sri_frame, durations=None):
        """ create a bar axes for the sri event plot

        Args:
            ax (matplotlib.pyplot.Axes):
            sri_frame (pandas.DataFrame): index=DatetimeIndex and columns=SRI
            durations (list):

        Returns:
            matplotlib.pyplot.Axes:
        """
        if durations is None:
            durations = [5, 10, 15, 20, 30, 45, 60, 90, 120, 180, 240, 360, 540, 720, 1080, 1440, 2880, 4320]

        # legend
        from matplotlib.lines import Line2D
        custom_lines = [Line2D([0], [0], color=self.indices_color[i], lw=4) for i in self.indices]
        names = [str(i) for i in self.indices]
        ax.legend(custom_lines, names, bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=len(self.indices),
                  mode="expand", borderaxespad=0., title='StarkRegenIndex', handlelength=0.7)

        duration_size = len(durations)
        # labels for the y axis
        durations_index = range(duration_size)
        dh = 1
        ax.set_yticks([i + dh / 2 for i in durations_index], minor=True)
        ax.set_yticks(list(durations_index), minor=False)

        ax.set_yticklabels(duration_steps_readable(durations), minor=True)
        ax.set_yticklabels([''] * duration_size, minor=False)
        ax.set_ylabel('duration of the design rainfall')

        # for the relative start time
        freq = guess_freq(sri_frame.index)
        start_period = sri_frame.index[0].to_period(freq).ordinal

        # idf_table.index = idf_table.index - idf_table.index[0]

        min_duration = pd.Timedelta(minutes=1)

        for hi, d in enumerate(sri_frame.columns):
            sri = sri_frame[d]

            for i in self.indices:
                # not really a rain event, but the results are the same
                tab = rain_events(sri, ignore_rain_below=i, min_gap=freq)

                if tab.empty:
                    continue

                if 1:
                    durations = (event_duration(tab) / min_duration).tolist()
                    rel_starts = ((tab[COL.START] - sri_frame.index[0]) / min_duration + start_period).tolist()
                    bar_x = list(zip(rel_starts, durations))
                else:
                    tab[COL.DUR] = event_duration(tab) / min_duration
                    bar_x = [(r[COL.START] / min_duration + start_period, r[COL.DUR]) for _, r in tab.iterrows()]

                ax.broken_barh(bar_x, (hi, dh), facecolors=self.indices_color[i])

        ax.set_ylim(0, duration_size)
        ax.set_xticklabels([])
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.axhline(0, color='k')
        ax.axhline(duration_size / 2, color='k')
        return ax

    @staticmethod
    def event_plot_caption(event, method, unit='mm'):
        """
        get a caption for the event

        Args:
            event (pd.Series | dict): statistics of the event
            method (str): used method for HRI estimation (i.e. SCHMITT, MUDERSBACH, KRUEGER_PFISTER)
            unit (str): unit of the observation

        Returns:
            str: caption for the plot
        """
        caption = event_caption(event, unit) + '\n'
        caption += f'The method used for the SRI calculation is: {method}.\n'

        if COL.MAX_SRI.format(method) in event:
            caption += f'The maximum SRI was {event[COL.MAX_SRI.format(method)]:0.2f}\n'

        if COL.MAX_SRI_DURATION.format(method) in event:
            caption += f'at a duration of {minutes_readable(event[COL.MAX_SRI_DURATION.format(method)])}.'
        return caption

    def event_plot_sri(self, event, durations=None, unit='mm', column_name='Precipitation'):
        """
        get a plot of the selected event

        Args:
            event (pandas.Series): event
            durations (list | numpy.ndarray): list of durations in minutes for the table
            unit (str): unit of the observation
            column_name (str): label of the observation

        Returns:
            (matplotlib.pyplot.Figure, matplotlib.pyplot.Axes): figure and axes of the plot
        """
        event = event.to_dict()
        start = event[COL.START]
        end = event[COL.END]

        plot_range = slice(start - pd.Timedelta(self._freq), end + pd.Timedelta(self._freq))

        if durations:
            max_dur = max(durations)
        else:
            max_dur = max(self.duration_steps)

        sri_frame_extended = self.get_sri_frame(
            self.series[start - pd.Timedelta(minutes=max_dur):
                        end + pd.Timedelta(self._freq)].asfreq(self._freq).fillna(0)
        )
        sri_frame = sri_frame_extended[plot_range]

        if COL.MAX_SRI.format(self.method) not in event:
            event[COL.MAX_SRI.format(self.method)] = sri_frame.max().max()
            event[COL.MAX_SRI_DURATION.format(self.method)] = sri_frame.max().idxmax()

        ts = self.series[plot_range].resample(self._freq).sum().fillna(0).copy()

        # -------------------------------------
        fig = plt.figure()

        sri_bar_ax = fig.add_subplot(211)
        sri_bar_ax = self.sri_bar_axes(sri_bar_ax, sri_frame_extended, durations)
        rain_ax = fig.add_subplot(212, sharex=sri_bar_ax)

        # -------------------------------------
        ts_sum, minutes = resample_rain_series(ts)
        rain_ax = rain_bar_plot(ts_sum, rain_ax)
        rain_ax.set_ylabel('{} in {}/{}min'.format(column_name, unit, minutes if minutes != 1 else ''))
        rain_ax.set_xlim(ts.index[0], ts.index[-1])

        return fig, self.event_plot_caption(event, self.method)

    def event_dataframe(self, event: dict) -> pandas.DataFrame:
        sri_table_event = pd.DataFrame(index=self.duration_steps)

        # self.method = self.METHODS.KRUEGER_PFISTER
        # sri_table_event[self.METHODS.KRUEGER_PFISTER] = self.get_event_sri_max(event[COL.START], event[COL.END]).astype(int)
        #
        # self.method = self.METHODS.MUDERSBACH
        # sri_table_event[self.METHODS.MUDERSBACH] = self.get_event_sri_max(event[COL.START], event[COL.END]).astype(int)

        # self.method = self.METHODS.SCHMITT
        sri_table_event[self.METHODS.SCHMITT] = self.get_event_sri_max(event[COL.START], event[COL.END]).astype(int)

        sri_table_event[COL.MAX_OVERLAPPING_SUM] = self.rainfall_sum_frame[event[COL.START]:event[COL.END]].max()
        sri_table_event[COL.MAX_PERIOD] = self.return_periods_frame[event[COL.START]:event[COL.END]].max()

        return sri_table_event.rename(minutes_readable)



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd
from matplotlib.ticker import NullFormatter

from .definitions import COL
from .little_helpers import duration_steps_readable
from .sww_utils import guess_freq, rain_events, event_duration

RETURN_PERIOD_COLORS = {
    # 0.5: '#e0ffff',  # 'lightcyan',
    1: '#00ffff',  # 'cyan',
    2: '#add8e6',  # 'lightblue',
    5: '#0000ff',  # 'blue',
    10: '#ffff00',  # 'yellow',
    20: '#ffa500',  # 'orange',
    50: '#ff0000',  # 'red',
    100: '#ff00ff',  # 'magenta',
}


def idf_bar_axes(ax, idf_table, return_period_colors=RETURN_PERIOD_COLORS):
    """
    create

    Args:
        ax (matplotlib.pyplot.Axes):
        idf_table (pandas.DataFrame):
        return_period_colors (dict): color of each return period {return period: color}

    Returns:
        matplotlib.pyplot.Axes:
    """
    return_periods = list(return_period_colors.keys())
    color_return_period = list(return_period_colors.values())

    # legend
    from matplotlib.lines import Line2D
    custom_lines = [Line2D([0], [0], color=c, lw=4) for c in color_return_period]
    names = ['{}a'.format(t) for t in return_periods]
    ax.legend(custom_lines, names, bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=len(color_return_period),
              mode="expand", borderaxespad=0., title='return periods')

    duration_size = len(idf_table.columns)
    # labels for the y axis
    durations_index = range(duration_size)
    dh = 1
    ax.set_yticks([i + dh/2 for i in durations_index], minor=True)
    ax.set_yticks(list(durations_index), minor=False)

    ax.set_yticklabels(duration_steps_readable(idf_table.columns), minor=True)
    ax.set_yticklabels([''] * duration_size, minor=False)
    ax.set_ylabel('duration of the design rainfall')

    # for the relative start time
    freq = guess_freq(idf_table.index)
    start_period = idf_table.index[0].to_period(freq).ordinal

    # idf_table.index = idf_table.index - idf_table.index[0]

    min_duration = pd.Timedelta(minutes=1)

    for hi, d in enumerate(idf_table.columns):
        tn = idf_table[d]

        for t in return_periods:
            c = return_period_colors[t]
            # not really a rain event, but the results are the same
            # tab2 = rain_events(tn, ignore_rain_below=t, min_gap=pd.Timedelta(minutes=d))
            tab = rain_events(tn, ignore_rain_below=t, min_gap=freq)
            # if tab.size != tab2.size:
            #     print()
            if tab.empty:
                continue

            if 1:
                durations = (event_duration(tab) / min_duration).tolist()
                rel_starts = ((tab[COL.START] - idf_table.index[0]) / min_duration + start_period).tolist()
                bar_x = list(zip(rel_starts, durations))
            else:
                tab[COL.DUR] = event_duration(tab) / min_duration
                bar_x = [(r[COL.START] / min_duration + start_period, r[COL.DUR]) for _, r in tab.iterrows()]

            ax.broken_barh(bar_x, (hi, dh), facecolors=c)

    ax.tick_params(axis='y', which='minor', length=0)
    ax.grid(axis='y', which='major')

    ax.set_ylim(0, duration_size)
    ax.set_xticklabels([])
    ax.xaxis.set_major_formatter(NullFormatter())
    # ax.axhline(0, color='k')
    ax.axhline(duration_size / 2, color='k')
    return ax



__author__ = "David Camhy, Markus Pichler"
__copyright__ = "Copyright 2018, University of Technology Graz"
__credits__ = ["David Camhy", "Markus Pichler"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "David Camhy, Markus Pichler"

from datetime import tzinfo

import numpy as np
import pandas as pd
import pytz
from pandas.tseries.frequencies import to_offset

from .definitions import COL


class TimezoneError(Exception):
    """Some Error With a Timezone"""


class IdfError(Exception):
    """Some Error Within this Package"""


def check_tz(timezone):
    """
    check the give timezone

    :param timezone:
    :type timezone: tzinfo | str
    :return:
    :rtype: tzinfo
    """
    if isinstance(timezone, str):
        try:
            return pytz.timezone(timezone)
        except:
            raise TimezoneError('"{}" is not a timezone or unknown'.format(timezone))
    # else:
    #     return timezone
    elif isinstance(timezone, tzinfo):
        return timezone
    else:
        raise TimezoneError('unknown timezone format: "{}"'.format(type(timezone)))


def remove_timezone(data):
    """
    convert the timezone to wintertime and then remove the timezone from the dataframe index, to not get in conflict
    with the plot functions

    Args:
        data (pandas.DataFrame | pandas.Series): data with datetimeindex 

    Returns:
        pandas.DataFrame | pandas.Series: data with datetimeindex without timezone
    """
    no_tz = data.copy()
    index = no_tz.index
    native_timezone = pytz.timezone('Etc/GMT-1')
    if index.tz is None:
        raise TimezoneError("Dataframe must have a timezone information")
    timezone = check_tz(native_timezone)
    idx = index.tz_convert(check_tz(timezone))
    no_tz.index = idx.tz_localize(None)
    return no_tz


########################################################################################################################
def guess_freq(date_time_index, default=pd.Timedelta(minutes=1)):
    """
    guess the frequency by evaluating the most often frequency

    Args:
        date_time_index (pandas.DatetimeIndex): index of a time-series
        default (pandas.Timedelta):

    Returns:
        pandas.DateOffset: frequency of the date-time-index
    """
    freq = date_time_index.freq
    if pd.notnull(freq):
        return to_offset(freq)

    if not len(date_time_index) <= 3:
        freq = pd.infer_freq(date_time_index)  # 'T'

        if pd.notnull(freq):
            return to_offset(freq)

        delta_series = date_time_index.to_series().diff(periods=1).fillna(method='backfill')
        counts = delta_series.value_counts()
        counts.drop(pd.Timedelta(minutes=0), errors='ignore')

        if counts.empty:
            delta = default
        else:
            delta = counts.index[0]
            if delta == pd.Timedelta(minutes=0):
                delta = default
    else:
        delta = default

    return to_offset(delta)


########################################################################################################################
def year_delta(years):
    return pd.Timedelta(days=365.25 * years)


########################################################################################################################
def rain_events(series, ignore_rain_below=0, min_gap=pd.Timedelta(hours=4)):
    """
    get rain events as a table with start and end times

    Args:
        series (pandas.Series): rain series
        ignore_rain_below (float): where it is considered as rain
        min_gap (pandas.Timedelta): 4 hours of no rain between events

    Returns:
        pandas.DataFrame: table of the rain events
    """
    # best OKOSTRA adjustment with 0.0
    # by ignoring 0.1 mm the results are getting bigger

    # remove values below a from the database
    temp = series[series > ignore_rain_below].index.to_series()

    # 4 hours of no rain between events

    event_end = temp[temp.diff(periods=-1) < -min_gap]
    event_end = event_end.append(temp.tail(1), ignore_index=True)

    event_start = temp[temp.diff() > min_gap]
    event_start = event_start.append(temp.head(1), ignore_index=True)
    event_start = event_start.sort_values().reset_index(drop=True)

    events = pd.concat([event_start, event_end], axis=1, ignore_index=True)
    events.columns = [COL.START, COL.END]
    return events


def event_number_to_series(events, index):
    """
    make a time-series where the value of the event number is paste to the <index>

    Args:
        events (pandas.DataFrame):
        index (pandas.DatetimeIndex):

    Returns:
        pandas.Series:
    """
    ts = pd.Series(index=index)

    events_dict = events.to_dict(orient='index')
    for event_no, event in events_dict.items():
        ts[event[COL.START]: event[COL.END]] = event_no

    return ts


########################################################################################################################
def agg_events(events, series, agg='sum'):
    """

    Args:
        events (pandas.DataFrame): table of events
        series (pandas.Series): time-series data
        agg (str | function): aggregation of time-series

    Returns:
        numpy.ndarray: result of function of every event
    """
    if events.empty:
        return np.array([])

    if events.index.size > 3500:
        res = series.groupby(event_number_to_series(events, series.index)).agg(agg).values
    else:
        res = events.apply(lambda event: series[event[COL.START]:event[COL.END]].agg(agg), axis=1).values
    return res


########################################################################################################################
def event_duration(events):
    """
    calculate the event duration

    Args:
        events (pandas.DataFrame): table of events with COL.START and COL.END times

    Returns:
        pandas.Series: duration of each event
    """
    return events[COL.END] - events[COL.START]


########################################################################################################################
def rain_bar_plot(rain, ax=None, color='#1E88E5', reverse=False, step='post', joinstyle='miter', capstyle='butt'):
    """
    make a standard precipitation/rain plot

    Args:
        rain (pandas.Series):
        ax (matplotlib.axes.Axes):
        color (str):
        reverse (bool):
        drawstyle (str):  'steps-mid' 'steps-post' steps-pre'

    Returns:
        matplotlib.axes.Axes: rain plot
    """
    ax = rain.plot(ax=ax, drawstyle=f'steps-{step}', color=color, solid_capstyle=capstyle, solid_joinstyle=joinstyle,
                   lw=0)
    ax.fill_between(rain.index, rain.values, 0, step=step, zorder=1000, color=color, capstyle=capstyle,
                    joinstyle=joinstyle)

    if reverse:
        # ax.set_ylim(top=0, bottom=rain.max() * 1.1)
        ax.set_ylim(bottom=0)
        ax.invert_yaxis()
    else:
        ax.set_ylim(bottom=0)

    return ax


########################################################################################################################
def resample_rain_series(series):
    """

    Args:
        series (pandas.Series):

    Returns:
        tuple[pandas.Series, int]: the resampled series AND the final frequency in minutes
    """
    resample_minutes = (
        (pd.Timedelta(hours=5), 1),
        (pd.Timedelta(hours=12), 2),
        (pd.Timedelta(days=1), 5),
        (pd.Timedelta(days=2), 10),
        (pd.Timedelta(days=3), 15),
        (pd.Timedelta(days=4), 20)
    )

    dur = series.index[-1] - series.index[0]
    freq = guess_freq(series.index)

    minutes = 1
    for duration_limit, minutes in resample_minutes:
        if dur < duration_limit:
            break

    if freq.delta > pd.Timedelta(minutes=minutes):
        return series, int(freq / pd.Timedelta(minutes=1))

    # print('resample_rain_series: ', dur, duration_limit, minutes)
    return series.resample('{}T'.format(minutes)).sum(), minutes
